use clap::Parser;
use regex::bytes::Regex;
use std::fmt::{Display, Formatter};
use std::fs::{ReadDir};
use std::ops::Range;
use std::path::PathBuf;
use std::{fs, thread};
use std::sync::atomic::{AtomicUsize, Ordering};
use std::sync::mpsc::{sync_channel, SyncSender};
use std::time::Duration;


static ATOMIC_THREADS: AtomicUsize = AtomicUsize::new(0);
static ATOMIC_SEARCH_CTR: AtomicUsize = AtomicUsize::new(0);


#[derive(Parser, Debug)]
#[command(author, version, about, long_about = None)]
struct Args {
    /// The file/folder name filter
    #[arg(short, long)]
    filter: Option<String>,

    /// The regex pattern that the user provided
    regex: String,

    /// The paths in which mygrep should search, if empty, in the current directory
    paths: Vec<String>,
}

/// get recursively next directory
pub fn get_dir<E>(dir: Result<ReadDir, E>, regex: Regex, tx: SyncSender<GrepResult>){
    match dir {
        Ok(dir) => {
            for j in dir{
                let entry = j;
                match entry {
                    Ok(entry) => {
                        /// call get file function that uses the path to check if there are files or directories
                        get_file(entry.path(), regex.clone(), tx.clone());
                    }
                    _=> println!("Error in entry getting of directory!"),
                }
            }
        },
        _=> println!("Error in directory!"),
    }
}

pub fn get_file(paths: PathBuf, regex: Regex, tx: SyncSender<GrepResult>){

    /// see all threads that is available for system
    let threads = usize::from(std::thread::available_parallelism().unwrap());

    /// if file make thread for every file
    if paths.is_file(){
        /// wait if all threads are active
        while ATOMIC_THREADS.load(Ordering::SeqCst)  == threads * 1000 {
            //thread::sleep(Duration::from_millis(1));        // system sleep
        }

        /// make threads and add one to ATOMIC_THREADS to make sure that only a limited amount of threads are active
        ATOMIC_THREADS.fetch_add(1, Ordering::SeqCst);
        let handle = thread::spawn(move || search(&paths, &regex, tx.clone()));
        let join = handle.join();
        match join {
            Ok(_join) => (),
            _=> println!("Error in joining threads"),
        }
    }
    else {
        ///get recursively next directory
        let dir = paths.read_dir();
        get_dir(dir, regex.clone(), tx.clone());
    }
}

pub fn search(paths: &PathBuf, regex: &Regex, tx: SyncSender<GrepResult>) {

    //println!("{:?}", ATOMIC_THREADS);

    let file = fs::read(paths);         // read data from file
    match file {
        Ok(file) => {
            let mut range_vec = Vec::new();        // create new vector for the found ranges
            let mut regex_found = false;                     // only fill and send result struct if there is a match

            /// loop over all the matches found in the file and add them to the vector and set the boolean for found regex on true
            for range in regex.find_iter(&*file){
                range_vec.push(Range{start: range.start(), end: range.end() });
                regex_found = true;
            }

            /// if match make result struct and send the data
            if regex_found{
                let result = GrepResult {
                    path: paths.clone(),
                    content: file,
                    ranges: range_vec,
                    search_ctr: ATOMIC_SEARCH_CTR.fetch_add(1, Ordering::SeqCst),   // add atomic because of multiple threads adding to same address
                };
                let send_error = tx.send(result);                    // send the result
                match send_error {
                    Ok(_send_error) => (),
                    _=> println!("Error in sending message!")
                }
            }
        }
        _ => { println!("Error reading file");}
    }
    /// subtract threads count
    ATOMIC_THREADS.fetch_sub(1, Ordering::SeqCst);
}

fn main() {
    ///Parse arguments, using the clap crate
    let args: Args = Args::parse();
    let regex = Regex::new(&args.regex).unwrap();

    /// Get the paths that we should search
    let paths = if args.paths.is_empty() {
        ///If no paths were provided, we search the current path
        vec![std::env::current_dir().unwrap()]
    } else {
        /// Take all paths from the command line arguments, and map the paths to create PathBufs
        args.paths.iter().map(PathBuf::from).collect()
    };


    let (tx, rx) = sync_channel(10);
    thread::spawn(move || {
        while let Ok(i) = rx.recv(){
            println!("{ }", i);
        }
    });


    /// loop through paths
    for i in paths.iter(){
        /// get directory
        let dir = i.read_dir();
        get_dir(dir, regex.clone(), tx.clone());
    }
}

/// This structure represents the matches that the tool found in **a single file**.
/// It implements `Display`, so it can be pretty-printed.
/// This struct and the `Display` trait implementation do NOT need to be edited.
pub struct GrepResult {
    /// the path of the search result
    path: PathBuf,

    /// the contents of that file
    content: Vec<u8>,

    /// which ranges in the file match the filter.
    /// A file may contain more than one match. Each match is a Range,
    /// which is a start and end byte offset in the content field.
    ranges: Vec<Range<usize>>,

    /// The index of this search result (ie. a counter of how many files have had a match before this
    /// one). Note that this count must always increase as the results are printed.
    search_ctr: usize,
}

impl Display for GrepResult {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        let max_context = 70;

        if self.ranges.is_empty() {
            return Ok(());
        }

        writeln!(f, ">>> (#{}) {:?}", self.search_ctr, self.path)?;
        for range in &self.ranges {
            /// Find the index of the first byte before the range that is a newline character, plus one.
            let mut ctx_start = (0..range.start)
                .rev()
                .find(|p| self.content[*p] == b'\n' || self.content[*p] == b'\r')
                .map(|p| p + 1)
                .unwrap_or(0);

            /// Find the index of the first byte after the range that is a newline character (not minus one because it is exclusive)
            let mut ctx_end = (range.end..self.content.len())
                .find(|p| self.content[*p] == b'\n' || self.content[*p] == b'\r')
                .unwrap_or(self.content.len());

            /// if the context is too large, reduce its size
            if ctx_start + max_context < range.start {
                ctx_start = range.start - max_context;
            }
            if ctx_end > range.end + max_context {
                ctx_end = range.end + max_context;
            }

            /// Finally, print the result
            writeln!(
                f,
                "{}",
                String::from_utf8_lossy(&self.content[ctx_start..ctx_end])
            )?;
            /// Print ^^^^ underneath matched part
            writeln!(
                f,
                "{}{}{}",
                " ".repeat(range.start - ctx_start),
                "^".repeat(range.end - range.start),
                " ".repeat(ctx_end - range.end)
            )?;
        }

        Ok(())
    }
}
